@@ -7,28 +7,42 @@
 
 '''pull, update and merge in one command (DEPRECATED)'''
 
-from mercurial.i18n import _
-from mercurial.node import short
-from mercurial import commands, cmdutil, hg, util, error
-from mercurial.lock import release
-from mercurial import exchange
+from __future__ import absolute_import
 
+from mercurial.i18n import _
+from mercurial.node import (
+    short,
+)
+from mercurial import (
+    cmdutil,
+    error,
+    exchange,
+    hg,
+    lock,
+    pycompat,
+    registrar,
+    util,
+)
+from mercurial.utils import dateutil
+
+release = lock.release
 cmdtable = {}
-command = cmdutil.command(cmdtable)
-# Note for extension authors: ONLY specify testedwith = 'internal' for
+command = registrar.command(cmdtable)
+# Note for extension authors: ONLY specify testedwith = 'ships-with-hg-core' for
 # extensions which SHIP WITH MERCURIAL. Non-mainline extensions should
 # be specifying the version(s) of Mercurial they are tested with, or
 # leave the attribute unspecified.
-testedwith = 'internal'
+testedwith = 'ships-with-hg-core'
 
 @command('fetch',
     [('r', 'rev', [],
      _('a specific revision you would like to pull'), _('REV')),
-    ('e', 'edit', None, _('invoke editor on commit messages')),
+    ('', 'edit', None, _('invoke editor on commit messages')),
     ('', 'force-editor', None, _('edit commit message (DEPRECATED)')),
     ('', 'switch-parent', None, _('switch parents when merging')),
-    ] + commands.commitopts + commands.commitopts2 + commands.remoteopts,
-    _('hg fetch [SOURCE]'))
+    ] + cmdutil.commitopts + cmdutil.commitopts2 + cmdutil.remoteopts,
+    _('hg fetch [SOURCE]'),
+    helpcategory=command.CATEGORY_REMOTE_REPO_MANAGEMENT)
 def fetch(ui, repo, source='default', **opts):
     '''pull changes from a remote repository, merge new changes if needed.
 
@@ -49,9 +63,10 @@ def fetch(ui, repo, source='default', **opts):
     Returns 0 on success.
     '''
 
+    opts = pycompat.byteskwargs(opts)
     date = opts.get('date')
     if date:
-        opts['date'] = util.parsedate(date)
+        opts['date'] = dateutil.parsedate(date)
 
     parent, _p2 = repo.dirstate.parents()
     branch = repo.dirstate.branch()
@@ -61,7 +76,7 @@ def fetch(ui, repo, source='default', **opts):
         branchnode = None
     if parent != branchnode:
         raise error.Abort(_('working directory not at branch tip'),
-                         hint=_('use "hg update" to check out branch tip'))
+                         hint=_("use 'hg update' to check out branch tip"))
 
     wlock = lock = None
     try:
