@@ -103,12 +103,19 @@ notify.maxdiff
   Maximum number of diff lines to include in notification email. Set to 0
   to disable the diff, or -1 to include all of it. Default: 300.
 
+notify.maxdiffstat
+  Maximum number of diffstat lines to include in notification email. Set to -1
+  to include all of it. Default: -1.
+
 notify.maxsubject
   Maximum number of characters in email's subject line. Default: 67.
 
 notify.diffstat
   Set to True to include a diffstat before diff content. Default: True.
 
+notify.showfunc
+  If set, override ``diff.showfunc`` for the diff content. Default: None.
+
 notify.merge
   If True, send notifications for merge changesets. Default: True.
 
@@ -134,7 +141,7 @@ web.baseurl
 '''
 from __future__ import absolute_import
 
-import email
+import email.errors as emailerrors
 import email.parser as emailparser
 import fnmatch
 import socket
@@ -142,13 +149,18 @@ import time
 
 from mercurial.i18n import _
 from mercurial import (
-    cmdutil,
+    encoding,
     error,
+    logcmdutil,
     mail,
     patch,
     registrar,
     util,
 )
+from mercurial.utils import (
+    dateutil,
+    stringutil,
+)
 
 # Note for extension authors: ONLY specify testedwith = 'ships-with-hg-core' for
 # extensions which SHIP WITH MERCURIAL. Non-mainline extensions should
@@ -180,6 +192,9 @@ configitem('notify', 'incoming',
 configitem('notify', 'maxdiff',
     default=300,
 )
+configitem('notify', 'maxdiffstat',
+    default=-1,
+)
 configitem('notify', 'maxsubject',
     default=67,
 )
@@ -195,6 +210,9 @@ configitem('notify', 'outgoing',
 configitem('notify', 'sources',
     default='serve',
 )
+configitem('notify', 'showfunc',
+    default=None,
+)
 configitem('notify', 'strip',
     default=0,
 )
@@ -209,7 +227,7 @@ configitem('notify', 'test',
 )
 
 # template for single changeset can include email headers.
-single_template = '''
+single_template = b'''
 Subject: changeset in {webroot}: {desc|firstline|strip}
 From: {author}
 
@@ -222,7 +240,7 @@ description:
 # template for multiple changesets should not contain email headers,
 # because only first set of headers will be used and result will look
 # strange.
-multiple_template = '''
+multiple_template = b'''
 changeset {node|short} in {root}
 details: {baseurl}{webroot}?cmd=changeset;node={node|short}
 summary: {desc|firstline}
@@ -249,6 +267,9 @@ class notifier(object):
         self.charsets = mail._charsets(self.ui)
         self.subs = self.subscribers()
         self.merge = self.ui.configbool('notify', 'merge')
+        self.showfunc = self.ui.configbool('notify', 'showfunc')
+        if self.showfunc is None:
+            self.showfunc = self.ui.configbool('diff', 'showfunc')
 
         mapfile = None
         template = (self.ui.config('notify', hooktype) or
@@ -257,9 +278,8 @@ class notifier(object):
             mapfile = self.ui.config('notify', 'style')
         if not mapfile and not template:
             template = deftemplates.get(hooktype) or single_template
-        spec = cmdutil.logtemplatespec(template, mapfile)
-        self.t = cmdutil.changeset_templater(self.ui, self.repo, spec,
-                                             False, None, False)
+        spec = logcmdutil.templatespec(template, mapfile)
+        self.t = logcmdutil.changesettemplater(self.ui, self.repo, spec)
 
     def strip(self, path):
         '''strip leading slashes from local path, turn into web-safe path.'''
@@ -277,7 +297,7 @@ class notifier(object):
     def fixmail(self, addr):
         '''try to clean up email addresses.'''
 
-        addr = util.email(addr.strip())
+        addr = stringutil.email(addr.strip())
         if self.domain:
             a = addr.find('@localhost')
             if a != -1:
@@ -342,13 +362,14 @@ class notifier(object):
 
         p = emailparser.Parser()
         try:
-            msg = p.parsestr(data)
-        except email.Errors.MessageParseError as inst:
+            msg = p.parsestr(encoding.strfromlocal(data))
+        except emailerrors.MessageParseError as inst:
             raise error.Abort(inst)
 
         # store sender and subject
-        sender, subject = msg['From'], msg['Subject']
-        del msg['From'], msg['Subject']
+        sender = encoding.strtolocal(msg[r'From'])
+        subject = encoding.strtolocal(msg[r'Subject'])
+        del msg[r'From'], msg[r'Subject']
 
         if not msg.is_multipart():
             # create fresh mime message from scratch
@@ -361,7 +382,8 @@ class notifier(object):
             for k, v in headers:
                 msg[k] = v
 
-        msg['Date'] = util.datestr(format="%a, %d %b %Y %H:%M:%S %1%2")
+        msg[r'Date'] = encoding.strfromlocal(
+            dateutil.datestr(format="%a, %d %b %Y %H:%M:%S %1%2"))
 
         # try to make subject line exist and be useful
         if not subject:
@@ -372,26 +394,27 @@ class notifier(object):
                 subject = '%s: %s' % (self.root, s)
         maxsubject = int(self.ui.config('notify', 'maxsubject'))
         if maxsubject:
-            subject = util.ellipsis(subject, maxsubject)
-        msg['Subject'] = mail.headencode(self.ui, subject,
-                                         self.charsets, self.test)
+            subject = stringutil.ellipsis(subject, maxsubject)
+        msg[r'Subject'] = encoding.strfromlocal(
+            mail.headencode(self.ui, subject, self.charsets, self.test))
 
         # try to make message have proper sender
         if not sender:
             sender = self.ui.config('email', 'from') or self.ui.username()
         if '@' not in sender or '@localhost' in sender:
             sender = self.fixmail(sender)
-        msg['From'] = mail.addressencode(self.ui, sender,
-                                         self.charsets, self.test)
+        msg[r'From'] = encoding.strfromlocal(
+            mail.addressencode(self.ui, sender, self.charsets, self.test))
 
-        msg['X-Hg-Notification'] = 'changeset %s' % ctx
-        if not msg['Message-Id']:
-            msg['Message-Id'] = ('<hg.%s.%s.%s@%s>' %
-                                 (ctx, int(time.time()),
-                                  hash(self.repo.root), socket.getfqdn()))
-        msg['To'] = ', '.join(sorted(subs))
+        msg[r'X-Hg-Notification'] = r'changeset %s' % ctx
+        if not msg[r'Message-Id']:
+            msg[r'Message-Id'] = encoding.strfromlocal(
+                '<hg.%s.%d.%d@%s>' % (ctx, int(time.time()),
+                                      hash(self.repo.root),
+                                      encoding.strtolocal(socket.getfqdn())))
+        msg[r'To'] = encoding.strfromlocal(', '.join(sorted(subs)))
 
-        msgtext = msg.as_string()
+        msgtext = encoding.strtolocal(msg.as_string())
         if self.test:
             self.ui.write(msgtext)
             if not msgtext.endswith('\n'):
@@ -399,7 +422,7 @@ class notifier(object):
         else:
             self.ui.status(_('notify: sending %d subscribers %d changes\n') %
                            (len(subs), count))
-            mail.sendmail(self.ui, util.email(msg['From']),
+            mail.sendmail(self.ui, stringutil.email(msg[r'From']),
                           subs, msgtext, mbox=self.mbox)
 
     def diff(self, ctx, ref=None):
@@ -410,15 +433,23 @@ class notifier(object):
             ref = ref.node()
         else:
             ref = ctx.node()
-        chunks = patch.diff(self.repo, prev, ref,
-                            opts=patch.diffallopts(self.ui))
+        diffopts = patch.diffallopts(self.ui)
+        diffopts.showfunc = self.showfunc
+        chunks = patch.diff(self.repo, prev, ref, opts=diffopts)
         difflines = ''.join(chunks).splitlines()
 
         if self.ui.configbool('notify', 'diffstat'):
+            maxdiffstat = int(self.ui.config('notify', 'maxdiffstat'))
             s = patch.diffstat(difflines)
             # s may be nil, don't include the header if it is
             if s:
-                self.ui.write(_('\ndiffstat:\n\n%s') % s)
+                if maxdiffstat >= 0 and s.count("\n") > maxdiffstat + 1:
+                    s = s.split("\n")
+                    msg = _('\ndiffstat (truncated from %d to %d lines):\n\n')
+                    self.ui.write(msg % (len(s) - 2, maxdiffstat))
+                    self.ui.write("\n".join(s[:maxdiffstat] + s[-2:]))
+                else:
+                    self.ui.write(_('\ndiffstat:\n\n%s') % s)
 
         if maxdiff == 0:
             return
@@ -438,7 +469,7 @@ def hook(ui, repo, hooktype, node=None, source=None, **kwargs):
     changegroup. else send one email per changeset.'''
 
     n = notifier(ui, repo, hooktype)
-    ctx = repo[node]
+    ctx = repo.unfiltered()[node]
 
     if not n.subs:
         ui.debug('notify: no subscribers to repository %s\n' % n.root)
@@ -452,8 +483,7 @@ def hook(ui, repo, hooktype, node=None, source=None, **kwargs):
     count = 0
     author = ''
     if hooktype == 'changegroup' or hooktype == 'outgoing':
-        start, end = ctx.rev(), len(repo)
-        for rev in xrange(start, end):
+        for rev in repo.changelog.revs(start=ctx.rev()):
             if n.node(repo[rev]):
                 count += 1
                 if not author:
@@ -465,7 +495,7 @@ def hook(ui, repo, hooktype, node=None, source=None, **kwargs):
                 ui.pushbuffer()
         if count:
             n.diff(ctx, repo['tip'])
-    else:
+    elif ctx.rev() in repo:
         if not n.node(ctx):
             ui.popbuffer()
             ui.note(_('notify: suppressing notification for merge %d:%s\n') %
